# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import os
import re
import sys
import logging
from string import Template
from itertools import chain
from collections import OrderedDict

if sys.version_info[0] == 2:
    from extensions import *  # pylint: disable=E0401,W0401
else:
    from dharma.core.extensions import *  # pylint: disable=W0401,W0614


class GenState:
    def __init__(self):
        self.leaf_mode = False
        self.leaf_trigger = 0


class String:
    """Generator class basic strings which need no further evaluation."""

    def __init__(self, value, parent):
        self.parent = parent
        self.value = value

    def generate(self, state):
        return self.value


class ValueXRef:
    """Generator class for +value+ cross references."""

    def __init__(self, value, parent):
        self.value = ("%s:%s" % (parent.namespace, value)) if ":" not in value else value
        self.parent = parent
        self.parent.value_xref[self.value] = None

    def generate(self, state):
        try:
            ref = self.parent.value_xref[self.value]
        except KeyError:
            logging.error("Value xref inconsistency in %s looking for %s", self.parent.ident, self.value)
            sys.exit(-1)
        return ref.generate(state)


class VariableXRef:
    """Generator class for !variable! cross references."""

    def __init__(self, value, parent):
        self.value = ("%s:%s" % (parent.namespace, value)) if ":" not in value else value
        self.parent = parent
        self.parent.variable_xref[self.value] = None

    def generate(self, state):
        try:
            ref = self.parent.variable_xref[self.value]
        except KeyError:
            logging.error("Variable xref inconsistency in %s looking for %s", self.parent.ident, self.value)
            sys.exit(-1)
        return ref.generate(state)


class ElementXRef:
    """Generator class for @value@ cross references."""

    def __init__(self, value, parent):
        self.value = ("%s:%s" % (parent.namespace, value)) if ":" not in value else value
        self.parent = parent
        self.parent.element_xref[self.value] = None

    def generate(self, state):
        try:
            ref = self.parent.element_xref[self.value]
        except KeyError:
            logging.error("Element xref inconsistency in %s looking for %s", self.parent.ident, self.value)
            sys.exit(-1)
        return ref.generate(state)


class DharmaObject(list):
    """Base object of which Dharma section classes inherit from."""

    def __init__(self, ident, machine):
        list.__init__(self)
        self.ident = "%s:%s" % (machine.namespace, ident)
        self.machine = machine
        self.value_xref = {}
        self.variable_xref = {}
        self.element_xref = {}
        self.namespace = machine.namespace
        self.lineno = machine.lineno

    def id(self):  # pylint: disable=invalid-name
        return "Line %d [%s]" % (self.lineno, self.namespace)

    def __hash__(self):
        return hash(self.ident)

    @staticmethod
    def eval(tokens, state):
        return "".join(token.generate(state) for token in tokens)


class DharmaValue(DharmaObject):
    """Dharma class which manages the |value| section of a grammar."""

    def __init__(self, ident, machine):
        DharmaObject.__init__(self, ident, machine)
        self.leaf = []
        self.leaf_path = []
        self.path_idents = set()
        self.minimized = None

    def n_xrefs(self, value):
        repeats, n = False, 0
        for t in value:
            if isinstance(t, ValueXRef):
                n += 1
                if t.value not in self.path_idents:
                    return False, None, None
            elif isinstance(t, MetaRepeat):
                repeats = True
                n += self.n_xrefs(t.repeat)[2]
        return True, repeats, max(1, min(n, 8))  # constrain within [1, 8]

    def append(self, value):
        list.append(self, value)
        for t in value:
            if isinstance(t, (MetaRepeat, ValueXRef)):
                return
        self.leaf.append(value)

    def generate(self, state):  # pylint: disable=too-many-branches
        if not state.leaf_mode:
            state.leaf_trigger += 1
            if state.leaf_trigger > DharmaConst.LEAF_TRIGGER:
                state.leaf_mode = True
        if not self:
            return ""
        if state.leaf_mode and self.leaf:
            value = random.choice(self.leaf)
        elif state.leaf_mode:  # favour non-repeating
            if self.minimized is None:
                n_refs_groups = {}
                have_non_repeats = False
                for v in self:
                    is_leaf_path, repeats, n_xrefs = self.n_xrefs(v)
                    if not is_leaf_path:
                        continue
                    if not repeats:
                        if not have_non_repeats:
                            n_refs_groups = {}
                            have_non_repeats = True
                    if not repeats or not have_non_repeats:
                        n_refs_groups.setdefault(n_xrefs, []).append(v)
                for _, v in sorted(n_refs_groups.items()):
                    self.minimized = v
                    break
                if not self.minimized:
                    logging.error("No path to leaf in force-leaf mode in value %s", self.ident)
                    sys.exit(-1)
            value = random.choice(self.minimized)
        else:
            value = random.choice(self)
        return self.eval(value, state)


class DharmaVariable(DharmaObject):
    """Dharma class which manages the |variable| section of a grammar."""

    def __init__(self, ident, machine):
        DharmaObject.__init__(self, ident, machine)
        self.var = ident
        self.count = 0
        self.default = ""

    def clear(self):
        self.count = 0
        self.default = ""

    def generate(self, state):
        """Return a random variable if any, otherwise create a new default variable."""
        if self.count >= random.randint(DharmaConst.VARIABLE_MIN, DharmaConst.VARIABLE_MAX):
            return "%s%d" % (self.var, random.randint(1, self.count))
        var = random.choice(self)
        prefix = self.eval(var[0], state)
        suffix = self.eval(var[1], state)
        self.count += 1
        element_name = "%s%d" % (self.var, self.count)
        self.default += "%s%s%s\n" % (prefix, element_name, suffix)
        return element_name


class DharmaVariance(DharmaObject):
    """Dharma class which manages the |variance| section of a grammar."""

    def generate(self, state):
        return self.eval(random.choice(self), state)


class DharmaMachine:  # pylint: disable=too-many-instance-attributes
    def __init__(self, prefix="", suffix="", template=""):
        self.section = None
        self.level = "top"
        self.namespace = ""
        self.lineno = 0
        self.current_obj = None
        self.value = {}
        self.variable = OrderedDict()
        self.variance = {}
        self.prefix = prefix
        self.suffix = suffix
        self.template = template
        self.consts_set = {}
        self.default_grammars = ["../grammars/common.dg"]
        self.grammar_level_registry = r"""^(
            (?P<comment>%%%).*|
            %const%\s*(?P<const>[A-Z_]+)\s*:=\s*(?P<value>.*)|
            %section%\s*:=\s*(?P<section>value|variable|variance)|
            (?P<ident>[a-zA-Z0-9_]+)\s*:=\s*|
            (?P<empty>\s*)|
            (\t|[ ]+)(?P<assign>.*)
        )$"""
        self.xref_registry = r"""(
            (?P<type>\+|!|@)(?P<xref>[a-zA-Z0-9:_]+)(?P=type)|
            %uri%\(\s*(?P<uri>.*?)\s*\)|
            %repeat%\(\s*(?P<repeat>.+?)\s*(,\s*"(?P<separator>.*?)")?\s*(,\s*(?P<nodups>nodups))?\s*\)|
            %block%\(\s*(?P<block>.*?)\s*\)|
            %range%\((?P<start>.+?)-(?P<end>.+?)\)|
            %choice%\(\s*(?P<choices>.+?)\s*\)
        )"""

    def process_settings(self, settings):
        """A lazy way of feeding Dharma with configuration settings."""
        logging.debug("Using configuration from: %s", settings.name)
        exec(compile(settings.read(), settings.name, 'exec'), globals(), locals())  # pylint: disable=exec-used

    def set_namespace(self, name):
        self.namespace = name
        self.lineno = 0

    def id(self):  # pylint: disable=invalid-name
        return "Line %d [%s]" % (self.lineno, self.namespace)

    def parse_line(self, line):
        self.lineno += 1
        m = re.match(self.grammar_level_registry, line, re.VERBOSE | re.IGNORECASE)
        if m is None:
            pass
        elif m.group("comment"):
            return
        elif m.group("const"):
            self.handle_const(*m.group("const", "value"))
            return
        elif m.group("section"):
            self.handle_empty_line()
            self.section = m.group("section").lower()
            return
        elif m.group("empty") is not None:
            self.handle_empty_line()
            return
        elif self.section is None:
            logging.error("%s: Non-empty line in void section", self.id())
            sys.exit(-1)
        elif self.level == "top":
            self.handle_top_level(m.group("ident"))
            return
        elif self.level == "assign":
            self.handle_assign_level(m.group("assign"))
            return
        logging.error("%s: Unhandled line", self.id())
        sys.exit(-1)

    def handle_const(self, const, value):
        if not hasattr(DharmaConst, const):
            logging.error("%s: Trying to set non-existent constant", self.id())
            sys.exit(-1)
        orig = self.consts_set.get(const)
        if value[0] == '"':
            assert value[-1] == '"'
            value = value[1:-1]
            setattr(DharmaConst, const, value)
        else:
            setattr(DharmaConst, const, float(value) if "." in value else int(value))
        if orig is not None and getattr(DharmaConst, const) != orig:
            logging.warning("%s: Overriding constant %s defined by previous grammar", self.id(), const)
        self.consts_set[const] = getattr(DharmaConst, const)

    def handle_empty_line(self):
        if self.current_obj is None:
            pass
        elif not self.current_obj:
            logging.error("%s: Empty assignment", self.id())
            sys.exit(-1)
        else:
            self.add_section_object()
        self.level = "top"
        self.current_obj = None

    def handle_top_level(self, ident):
        if ident is None:
            logging.error("%s: Top level syntax error", self.id())
            sys.exit(-1)
        try:
            assign_type = {"value": DharmaValue,
                           "variable": DharmaVariable,
                           "variance": DharmaVariance}[self.section]
        except KeyError:
            logging.error("%s: Invalid state for top-level", self.id())
            sys.exit(-1)
        self.current_obj = assign_type(ident, self)
        self.level = "assign"

    def handle_assign_level(self, assign):
        if assign is None:
            logging.error("%s: Assign level syntax error", self.id())
            sys.exit(-1)
        try:
            parse_assign = {"value": self.parse_assign_value,
                            "variable": self.parse_assign_variable,
                            "variance": self.parse_assign_variance}[self.section]
        except KeyError:
            logging.error("%s: Invalid state for assignment", self.id())
            sys.exit(-1)
        parse_assign(self.parse_xrefs(assign))

    def parse_xrefs(self, token):
        """Search token for +value+ and !variable! style references. Be careful to not xref a new variable.
        """
        out, end = [], 0
        token = token.replace("\\n", "\n")
        for m in re.finditer(self.xref_registry, token, re.VERBOSE | re.DOTALL):
            if m.start(0) > end:
                out.append(String(token[end:m.start(0)], self.current_obj))
            end = m.end(0)
            if m.group("type"):
                xref_type = {"+": ValueXRef,
                             "!": VariableXRef,
                             "@": ElementXRef}[m.group("type")]
                out.append(xref_type(m.group("xref"), self.current_obj))
            elif m.group("uri") is not None:
                path = m.group("uri")
                out.append(MetaURI(path, self.current_obj))
            elif m.group("repeat") is not None:
                repeat, separator, nodups = m.group("repeat", "separator", "nodups")
                if separator is None:
                    separator = ""
                if nodups is None:
                    nodups = ""
                out.append(MetaRepeat(self.parse_xrefs(repeat), separator, nodups, self.current_obj))
            elif m.group("block") is not None:
                path = m.group("block")
                out.append(MetaBlock(path, self.current_obj))
            elif m.group("choices") is not None:
                choices = m.group("choices")
                out.append(MetaChoice(choices, self.current_obj))
            else:
                startval, endval = m.group("start", "end")
                out.append(MetaRange(startval, endval, self.current_obj))
        if end < len(token):
            out.append(String(token[end:], self.current_obj))
        return out

    def parse_assign_value(self, tokens):
        if not isinstance(self.current_obj, DharmaValue):
            logging.error("%s: Normal value found in non-normal assignment", self.id())
            sys.exit(-1)
        self.current_obj.append(tokens)

    def parse_assign_variable(self, tokens):
        """
        Example:
            tokens
                dharma.String:      'let ',
                dharma.ElementXRef: 'GrammarNS:<VarName>',
                dharma.String:      '= new ',
                dharma.ValueXRef:   'GrammarNS:<ValueName>'
        """
        for i, token in enumerate(tokens):
            if isinstance(token, ElementXRef):
                variable = token.value
                break
        else:
            logging.error("%s: Variable assignment syntax error", self.id())
            sys.exit(-1)
        if variable != self.current_obj.ident:
            logging.error("%s: Variable name mismatch", self.id())
            sys.exit(-1)
        if not isinstance(self.current_obj, DharmaVariable):
            logging.error("%s: Inconsistent object for variable assignment", self.id())
            sys.exit(-1)
        prefix, suffix = tokens[:i], tokens[i + 1:]  # pylint: disable=undefined-loop-variable
        self.current_obj.append((prefix, suffix))

    def parse_assign_variance(self, tokens):
        if not isinstance(self.current_obj, DharmaVariance):
            logging.error("%s: Inconsistent object for variance assignment", self.id())
            sys.exit(-1)
        self.current_obj.append(tokens)

    def add_section_object(self):
        try:
            section_dict = getattr(self, self.section)
        except AttributeError:
            logging.error("%s: Inconsistent section value, fatal", self.id())
            sys.exit(-1)
        if self.current_obj.ident in section_dict:
            logging.error("%s(%s): '%s' gets redefined", self.id(), self.section, self.current_obj.ident)
            sys.exit(-1)
        section_dict[self.current_obj.ident] = self.current_obj

    def resolve_xref(self):
        for obj in chain(self.value.values(),
                         self.variable.values(),
                         self.variance.values()):
            try:
                msg = "%s: Undefined value reference from %s to %s"
                obj.value_xref.update((x, self.value[x]) for x in obj.value_xref)
                msg = "%s: Undefined variable reference from %s to %s"
                obj.variable_xref.update((x, self.variable[x]) for x in obj.variable_xref)
                msg = "%s: Element reference without a default variable from %s to %s"
                obj.element_xref.update((x, self.variable[x]) for x in obj.element_xref)
            except KeyError as error:
                logging.error(msg, self.id(), obj.ident, error.args[0])
                sys.exit(-1)

    def calculate_leaf_paths(self):
        """Build map of reverse xrefs then traverse backwards marking path to leaf for all leaves.
        """
        reverse_xref = {}
        leaves = set()
        for v in self.value.values():
            if v.leaf:
                leaves.add(v)
            for xref in v.value_xref:
                reverse_xref.setdefault(xref, []).append(v.ident)
        for leaf in leaves:
            self.calculate_leaf_path(leaf, reverse_xref)

    def calculate_leaf_path(self, leaf, reverse_xref):
        if leaf.ident not in reverse_xref:
            return
        for name in reverse_xref[leaf.ident]:
            xref = self.value[name]
            xref.leaf_path.append((leaf.ident, leaf.ident, 0))
            xref.path_idents.add(leaf.ident)
            self.propagate_leaf(leaf.ident, xref, {xref}, 1, reverse_xref)

    def propagate_leaf(self, leaf, obj, node_seen, depth, reverse_xref):
        if obj.ident not in reverse_xref:
            return
        for name in reverse_xref[obj.ident]:
            xref = self.value[name]
            xref.leaf_path.append((leaf, obj.ident, depth))
            xref.path_idents.add(obj.ident)
            if xref in node_seen:
                continue
            node_seen.add(xref)
            self.propagate_leaf(leaf, xref, node_seen, depth + 1, reverse_xref)

    def generate_content(self):
        """Generates a test case as a string."""
        # Setup pre-conditions.
        if not self.variance:
            logging.error("%s: No variance information %s", self.id(), self.variance)
            sys.exit(-1)

        for var in self.variable.values():
            var.clear()

        # Handle variances
        variances = []
        for _ in range(random.randint(DharmaConst.VARIANCE_MIN, DharmaConst.VARIANCE_MAX)):
            var = random.choice(list(self.variance.values()))
            variances.append(DharmaConst.VARIANCE_TEMPLATE % var.generate(GenState()))
            variances.append("\n")

        # Handle variables
        variables = []
        for var in self.variable.values():
            if var.default:
                variables.append(DharmaConst.VARIANCE_TEMPLATE % var.default)
                variables.append("\n")

        # Build content
        content = "".join(chain([self.prefix], variables, variances, [self.suffix]))
        if self.template:
            return Template(self.template).safe_substitute(testcase_content=content)
        return content

    def generate_testcases(self, path, filetype, count):
        """Writes out generated test cases to the provided path."""
        path = path.rstrip("/")
        try:
            os.makedirs(path, exist_ok=True)
        except OSError as error:
            logging.error("Unable to create folder for test cases: %s", error)
            sys.exit(-1)
        for n in range(count):
            filename = os.path.join(path, "%d.%s" % (n + 1, filetype))
            try:
                with open(filename, "w") as fo:
                    fo.write(self.generate_content())
            except IOError:
                logging.error("Failed in writing test case %s", filename)
                sys.exit(-1)

    def process_grammars(self, grammars):
        """Process provided grammars by parsing them into Python objects."""
        for path in self.default_grammars:
            grammars.insert(0, open(os.path.relpath(os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                                                 os.path.normcase(path)))))
        for fo in grammars:
            logging.debug("Processing grammar content of %s", fo.name)
            self.set_namespace(os.path.splitext(os.path.basename(fo.name))[0])
            for line in fo:
                self.parse_line(line)
            self.handle_empty_line()
        self.resolve_xref()
        self.calculate_leaf_paths()



# pylint: disable = undefined-variable
DharmaConst.VARIANCE_MIN = 1
DharmaConst.VARIANCE_MAX = 8
DharmaConst.VARIABLE_MIN = 1
DharmaConst.VARIABLE_MAX = 4
DharmaConst.VARIANCE_TEMPLATE = "%s"
DharmaConst.MAX_REPEAT_POWER = 12
DharmaConst.LEAF_TRIGGER = 256
DharmaConst.URI_TABLE = {
    "images": "fuzzdata/samples/jpg/",
    "videos": "fuzzdata/samples/mp4/",
    "audios": "fuzzdata/samples/mp3/",
}



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import os
import sys
import math
import random
import logging


class DharmaConst:
    """Configuration settings for the Dharma generator."""
    URI_TABLE = {}
    LEAF_TRIGGER = 256
    MAX_REPEAT_POWER = 12
    VARIANCE_TEMPLATE = "%s"
    VARIANCE_MIN = 1
    VARIANCE_MAX = 8
    VARIABLE_MIN = 1
    VARIABLE_MAX = 4


class MetaBlock:
    """Grammar extension which loads code fragments from a file into the grammar."""

    def __init__(self, path, parent):
        self.parent = parent
        path = os.path.expanduser(path)
        if os.path.exists(path):
            with open(path) as fo:
                self.content = fo.read()
        else:
            logging.warning('%s: Unable to load resource for block() "%s"', parent.id(), path)
            self.content = path

    def generate(self, state):
        return self.content


class MetaURI:
    """Grammar extension which loads a random file URI into the generated code."""

    def __init__(self, path, parent):
        self.parent = parent
        if path in DharmaConst.URI_TABLE:
            path = DharmaConst.URI_TABLE[path]
        path = os.path.expanduser(path)
        if os.path.isdir(path):
            self.path = [p for p in (os.path.join(path, f) for f in os.listdir(path)) if os.path.isfile(p)]
        elif os.path.exists(path):
            self.path = [path]
        else:
            logging.warning('%s: Unable to identify argument of uri() "%s"', parent.id(), path)
            self.path = [path]

    def generate(self, state):
        return random.choice(self.path)


class MetaRepeat:
    """Grammar extension method which repeats an arbitrary expression."""

    def __init__(self, repeat, separator, nodups, parent):
        self.parent = parent
        self.repeat, self.separator, self.nodups = repeat, separator, nodups

    def generate(self, state):
        count = random.randint(1, math.pow(2, random.randint(1, DharmaConst.MAX_REPEAT_POWER)))
        strings = [self.parent.eval(self.repeat, state) for _ in range(count)]
        if self.nodups:
            strings = list(set(strings))
        return self.separator.join(strings)


class MetaChoice:
    """Grammar extension method which chooses an item out of a list randomly."""

    def __init__(self, choices, parent):
        self.parent = parent
        self.choices = choices
        self.choices = [x.strip() for x in self.choices.split(",")]

    def generate(self, state):
        return random.choice(self.choices)


class MetaRange:
    """Grammar extension method which generates a random value between a range of values |a| and |b|."""

    def __init__(self, a, b, parent):
        self.parent = parent
        self.base = None
        # Type identification
        if a is None or b is None:
            logging.error("%s: Malformed 'range' meta", parent.id())
            sys.exit(-1)
        if self._is_char(a) and self._is_char(b):
            self.a, self.b = ord(a), ord(b)
            self.fmt = "c"
            return
        if self._is_float(a) and self._is_float(b):
            type_a, type_b = float, float
            self.fmt = "f"
        else:
            type_a, type_b = int, int
            self.fmt = "i"
            if self._is_hex(a) and self._is_hex(b):
                self.base = 16
            else:
                self.base = 0
        # Type verification
        if type_a != type_b:
            logging.error("%s: Mismatch in 'range' meta %s/%s in %s",
                          parent.id(), type_a.__name__, type_b.__name__, parent.ident)
            sys.exit(-1)
        # Type construction
        try:
            if self.base:
                self.a, self.b = type_a(a, self.base), type_b(b, self.base)
            else:
                self.a, self.b = type_a(a), type_b(b)
        except ValueError:
            logging.error("%d: Conversion error %s in 'range' meta", parent.id(), type_b.__name__)
            sys.exit(-1)

    def _is_char(self, x):
        return len(x) == 1

    def _is_float(self, x):
        return "." in x

    def _is_hex(self, x):
        return "0x" in x

    def generate(self, state):
        if self.fmt == "c":
            return "%c" % random.randint(self.a, self.b)
        if self.fmt == "f":
            return "%g" % random.uniform(self.a, self.b)
        if self.fmt == "i":
            if self.base == 16:
                return "%x" % random.randint(self.a, self.b)
            return "%d" % random.randint(self.a, self.b)
        return None



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import base64
import email.message
import email.parser
import hashlib
import logging
import socket
import struct
import sys
import json

try:
    # python 3
    from socketserver import BaseRequestHandler, TCPServer, ThreadingMixIn
except ImportError:
    # python 2
    from SocketServer import BaseRequestHandler, TCPServer, ThreadingMixIn


class BaseWebSocketHandler(BaseRequestHandler):
    """Base class for WebSocket server."""

    REQUEST_TIMEOUT = 0.01
    _OPCODES = {
        0: 'continue',
        1: 'text',
        2: 'binary',
        8: 'close',
        9: 'ping',
        10: 'pong'
    }

    def handle(self):  # pylint: disable=too-many-branches,too-many-statements
        self.request.settimeout(self.REQUEST_TIMEOUT)
        str_t = str if sys.version_info[0] == 3 else lambda a, b: str(a).encode(b)
        while not self.should_close():
            try:
                _, headers = str_t(self.request.recv(1024), 'ascii').split('\r\n', 1)
                break
            except socket.timeout:
                # time.sleep(0.01)
                # print 'timeout 29'
                continue
        headers = email.parser.HeaderParser().parsestr(headers)
        # TODO(jschwartzentruber): validate request/headers
        hresponse = hashlib.sha1(headers['sec-websocket-key'].encode('ascii'))
        hresponse.update(b'258EAFA5-E914-47DA-95CA-C5AB0DC85B11')
        resp = email.message.Message()
        resp.add_header('Upgrade', 'websocket')
        resp.add_header('Connection', 'Upgrade')
        resp.add_header('Sec-WebSocket-Accept', str_t(base64.b64encode(hresponse.digest()), 'ascii'))
        resp = resp.as_string(unixfrom=False).replace('\n', '\r\n')
        self.request.sendall('HTTP/1.1 101 Switching Protocols\r\n{}'.format(resp).encode('ascii'))
        self.open()
        buf = None
        buf_op = None
        try:
            while not self.should_close():
                try:
                    data = struct.unpack('BB', self.request.recv(2))
                except socket.timeout:
                    # no data
                    # time.sleep(0.01)
                    # print 'timeout 51'
                    continue
                except struct.error:
                    break  # chrome doesn't send a close-frame
                fin, mask = bool(data[0] & 0x80), bool(data[1] & 0x80)
                opcode = self._OPCODES[data[0] & 0xF]
                if opcode == 'close':
                    break
                if opcode == 'pong':
                    self.on_pong()
                    continue
                length = data[1] & 0x7F
                if length == 126:
                    length = struct.unpack('!H', self.request.recv(2))[0]
                elif length == 127:
                    length = struct.unpack('!Q', self.request.recv(8))[0]
                mask = bytearray(self.request.recv(4)) if mask else None
                data = bytearray(self.request.recv(length))
                if mask is not None:
                    data = bytearray((b ^ mask[i % 4]) for (i, b) in enumerate(data))
                if opcode == 'continue':
                    assert buf is not None
                    opcode = buf_op
                elif opcode == 'ping':
                    self._send(10, data)
                    continue
                elif buf is not None:
                    logging.warning('Received a new frame while waiting for another to finish, '
                                    'discarding %u bytes of %u', len(buf), buf_op)
                    buf = buf_op = None
                if opcode == 'text':
                    data = str_t(data, 'utf8')
                elif opcode != 'binary':
                    logging.warning('Unknown websocket opcode "%s"', opcode)
                    continue
                if buf is None:
                    buf = data
                    buf_op = opcode
                else:
                    buf += data
                if fin:
                    self.on_message(buf)
                    buf = buf_op = None
        finally:
            self.on_close()

    def finish(self):
        pass

    def _send(self, opcode, data):
        length = len(data)
        out = bytearray()
        out.append(0x80 | opcode)
        if length <= 125:
            out.append(length)
        elif length <= 65535:
            out.append(126)
            out.extend(struct.pack('!H', length))
        else:
            out.append(127)
            out.extend(struct.pack('!Q', length))
        if length:
            out.extend(data)
        self.request.sendall(out)

    # Below is the partial API from tornado.websocket.WebSocketHandler
    def ping(self):
        self._send(9, '')

    def should_close(self):
        """When this returns true, the message loop will exit."""
        return False

    def write_message(self, message, binary=False):
        if binary:
            self._send(2, message)
        else:
            self._send(1, message.encode('utf8'))

    # Event handlers
    def on_pong(self):
        pass

    def open(self):
        pass

    def on_close(self):
        pass

    def on_message(self, message):
        raise NotImplementedError('Required method on_message() not implemented.')


class DharmaTCPServer(ThreadingMixIn, TCPServer):
    daemon_threads = True
    allow_reuse_address = True


class DharmaWebSocketServer:
    def __init__(self, machine, address=("127.0.0.1", 9090)):
        self.server = None
        self.machine = machine
        self.address = address

    def start(self):
        machine = self.machine

        class DharmaWebSocketHandler(BaseWebSocketHandler):
            def on_message(self, message):
                msg = json.loads(message)
                if msg.get("status") == "open":
                    logging.info("WebSocket connection opened.")
                if msg.get("status") in ("open", "success"):
                    self.write_message(machine.generate_content())
                elif msg.get("status") == "closed":
                    logging.info("WebSocket connection closed.")
                else:
                    logging.error("WebSocket received unexpected message %r", msg)

        try:
            self.server = DharmaTCPServer(self.address, DharmaWebSocketHandler)
        except Exception as error:  # pylint: disable=broad-except
            logging.error("Unable to start WebSocket server: %s", error)
            return
        logging.info("Socket server is listening at %s:%d", *self.address)
        self.server.serve_forever()

    def stop(self):
        if self.server is None:
            return
        try:
            logging.info("Stopping WebSocket server.")
            self.server.shutdown()
        except Exception as error:  # pylint: disable=broad-except
            logging.error("Unable to shutdown WebSocket server: %s", error)



#!/usr/bin/env python3
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import argparse
import logging
import os
import random
import struct
import sys

from .__version__ import __version__, __title__
from .core.dharma import DharmaMachine
from .core.websocket import DharmaWebSocketServer


class DharmaCommandLine:
    @staticmethod
    def parse_args():
        parser = argparse.ArgumentParser(
            add_help=False,
            description='Dharma Runtime',
            epilog='The exit status is 0 for non-failures and -1 for failures.',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            prog=__title__
        )

        m = parser.add_argument_group('mandatory arguments')
        m.add_argument('-grammars', metavar='file', type=argparse.FileType(), nargs='+', required=True,
                       help='input grammars')

        o = parser.add_argument_group('optional arguments')
        o.add_argument('-count', metavar='#', type=int, default=1, help='number of test cases')
        o.add_argument('-format', metavar='ext', default='html', help='format of test cases')
        o.add_argument('-h', '-help', '--help', action='help', help=argparse.SUPPRESS)
        o.add_argument('-logging', metavar='#', default=10, type=int, choices=range(10, 60, 10),
                       help='verbosity level of logging')
        o.add_argument('-prefix', metavar='file', type=argparse.FileType(), help='prefix data')
        o.add_argument('-recursion-limit', metavar='#', type=int, default=20000,
                       help='max python recursion limit')
        o.add_argument('-seed', metavar='#', type=int,
                       help='seed value for random, os.urandom will be used if not specified')
        o.add_argument('-server', action='store_true', help='run in server mode')
        o.add_argument('-server-host', metavar='host', type=str, default='127.0.0.1', help='server address')
        o.add_argument('-server-port', metavar='#', type=int, default=9090, help='server port')
        o.add_argument('-settings', metavar='file', type=argparse.FileType(),
                       default=os.path.join(os.path.dirname(os.path.abspath(__file__)), 'settings.py'),
                       help='')
        o.add_argument('-storage', metavar='path', help='folder for test cases')
        o.add_argument('-suffix', metavar='file', type=argparse.FileType(), help='suffix data')
        o.add_argument('-template', metavar='file', type=argparse.FileType(), help='template data')
        o.add_argument('-version', action='version', version='%(prog)s {}'.format(__version__),
                       help=argparse.SUPPRESS)

        return parser.parse_args()

    @classmethod
    def main(cls):
        args = cls.parse_args()
        sys.setrecursionlimit(args.recursion_limit)
        logging.basicConfig(format='[Dharma] %(asctime)s %(levelname)s: %(message)s', level=args.logging)
        if args.seed is None:
            args.seed = struct.unpack('q', os.urandom(8))[0]
        random.seed(args.seed)
        logging.info('Machine random seed: %d', args.seed)
        prefix_data = '' if not args.prefix else args.prefix.read()
        suffix_data = '' if not args.suffix else args.suffix.read()
        template_data = '' if not args.template else args.template.read()
        dharma = DharmaMachine(prefix_data, suffix_data, template_data)
        dharma.process_settings(args.settings)
        dharma.process_grammars(args.grammars)
        if args.storage:
            dharma.generate_testcases(args.storage, args.format, args.count)
        elif args.server:
            server = DharmaWebSocketServer(dharma, (args.server_host, args.server_port))
            try:
                server.start()
            except KeyboardInterrupt:
                pass
            finally:
                server.stop()
        else:
            for _ in range(args.count):
                print(dharma.generate_content())
        return 0



#!/usr/bin/env python3
# coding=utf-8
"""setuptools install script"""
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import os
import sys

from setuptools import setup, find_packages, Command


HERE = os.path.abspath(os.path.dirname(__file__))

ABOUT = {}
with open(os.path.join(HERE, 'dharma', '__version__.py')) as fo:
    exec(fo.read(), ABOUT)  # pylint: disable=exec-used


def README():
    with open("README.md") as fo:
        return fo.read()


class PublishCommand(Command):
    """Command class for: setup.py publish"""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(text):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(text))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            os.system('rm -rf {}'.format(os.path.join(HERE, 'dist')))
        except OSError:
            pass

        self.status('Building Source and Wheel distribution ...')
        os.system('{0} setup.py sdist bdist_wheel'.format(sys.executable))

        self.status('Pushing git tags ...')
        os.system('git tag v{0}'.format(ABOUT['__version__']))
        os.system('git push origin v{0}'.format(ABOUT['__version__']))

        self.status('Uploading the package to PyPI via Twine ...')
        os.system('twine upload dist/*')

        sys.exit()


if __name__ == "__main__":
    setup(
        version=ABOUT['__version__'],
        name=ABOUT['__title__'],
        license=ABOUT['__license__'],
        keywords=ABOUT['__keywords__'],
        description=ABOUT['__description__'],
        long_description=README(),
        long_description_content_type='text/markdown',
        author=ABOUT['__author__'],
        author_email=ABOUT['__author_email__'],
        maintainer=ABOUT['__maintainer__'],
        maintainer_email=ABOUT['__maintainer_email__'],
        url=ABOUT['__url__'],
        project_urls=ABOUT['__project_urls__'],
        download_url=ABOUT['__download_url__'],
        classifiers=[
            'Topic :: Software Development :: Testing',
            "Topic :: Security",
            'Intended Audience :: Developers',
            'Operating System :: OS Independent',
            "Programming Language :: Python :: 3",
            'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)'
        ],
        entry_points={
            "console_scripts": [
                "dharma = dharma:DharmaCommandLine.main"
            ]
        },
        packages=find_packages(),
        package_data={
            'dharma': [
                'grammars/*',
                'grammars/**/*'
            ]
        },
        cmdclass={
            'publish': PublishCommand
        }
    )



__title__ = 'dharma'
__version__ = '1.3.2'
__license__ = 'MPL 2.0'
__description__ = 'Generation-based, context-free grammar fuzzer.'
__keywords__ = 'fuzzer fuzzing security testing generator'
__author__ = 'Christoph Diehl'
__author_email__ = 'cdiehl@mozilla.com'
__maintainer__ = 'Mozilla Fuzzing Team'
__maintainer_email__ = 'fuzzing@mozilla.com'
__url__ = 'https://github.com/MozillaSecurity/dharma'
__project_urls__ = {
    'Bug Tracker': 'https://github.com/MozillaSecurity/dharma/issues'
}
__download_url__ = "https://github.com/MozillaSecurity/dharma/tarball/" + __version__
